// CubesView.cpp : implementation of the CCubesView class
//

#include "stdafx.h"
#include "Cubes.h"

#include "CubesDoc.h"
#include "CubesView.h"
#include "SceneTemplate.h"
#include "VisualObject.h"
#include "InfoViewManager.h"
#include "log.h"
#include "SocketCommand.h"
#include "CommandQueue.h"
#include "Function.h"

CommandQueue cq;
CommandQueue cq2;
FunctionManager functionManager;


static char* imageFile = NULL;
enum { SlowAsMud, SuperFast };
int timerMode = SlowAsMud;
static char* lastLabel = NULL;
static char* lastFileName = NULL;
char* transformName = NULL;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CCubesView

IMPLEMENT_DYNCREATE(CCubesView, CView)

BEGIN_MESSAGE_MAP(CCubesView, CView)
    //{{AFX_MSG_MAP(CCubesView)
    ON_WM_SIZE()
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_ADD_CUBE, OnAddCube)
	ON_COMMAND(ID_REMOVE_CUBE, OnRemoveCube)
	ON_COMMAND(ID_MOVE_UP, OnMoveUp)
	ON_COMMAND(ID_MOVE_DOWN, OnMoveDown)
	ON_COMMAND(ID_MOVE_RIGHT, OnMoveRight)
	ON_COMMAND(ID_MOVE_LEFT, OnMoveLeft)
	ON_COMMAND(ID_MOVE_FORWARD, OnMoveForward)
	ON_COMMAND(ID_MOVE_BACKWARD, OnMoveBackward)
	ON_COMMAND(ID_CHANGE_COLOR, OnChangeColor)
	ON_COMMAND(ID_BUTTON32781, OnTaller)
	ON_COMMAND(ID_BUTTON32782, OnShorter)
	ON_COMMAND(ID_BUTTON32783, OnFatter)
	ON_COMMAND(ID_BUTTON32784, OnThinner)
	ON_COMMAND(ID_BUTTON32788, OnFadeOut)
	ON_COMMAND(ID_BUTTON32789, OnFadeIn)
	ON_COMMAND(ID_BUTTON32785, OnCycle)
	ON_UPDATE_COMMAND_UI(ID_ADD_CUBE, OnUpdateAddCube)
	ON_WM_TIMER()
	ON_COMMAND(ID_BUTTON32792, OnRotateLeft)
	ON_COMMAND(ID_BUTTON32793, OnRotateRight)
	ON_COMMAND(ID_BUTTON32794, OnRotateTowards)
	ON_COMMAND(ID_BUTTON32795, OnRotateAway)
	ON_COMMAND(ID_BUTTON32796, OnTwistLeft)
	ON_COMMAND(ID_BUTTON32797, OnTwistRight)
	ON_COMMAND(ID_BUTTON32798, OnJiggle)
	ON_COMMAND(ID_BUTTON32799, OnMoveToMarker)
	//}}AFX_MSG_MAP
	ON_UPDATE_COMMAND_UI_RANGE(ID_REMOVE_CUBE, ID_BUTTON32799, OnUpdateCubeInScene)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCubesView construction/destruction


// holds templates we merge visuals with
static SceneTemplateManager stm;

// Holds info associated with visualizations.  This can be merged into templates
// if it exists and the template refers to it. 
//
// First use of this, "setinfo name value" is used to set values in here.
// Then in the "transform" template, it has a "%s" for the transform text
// and a "VAR_<name>" for the variable to be substituted.
//

static SceneTemplateManager svars;  
static InfoViewManager imv;


#include "VizState.h"
//VizState vizState;
void saveSession( const char* fileName ) {
	//vizState.saveSession( fileName );
	FILE* fp = fopen( fileName, "w" );
	fprintf( fp, "2\n" );
	if ( fp != NULL ) {
		imv.saveState( fp );
		fclose( fp );
	}
}

void openStateFile( const char* fileName ) {
	FILE* fp = fopen( fileName, "r" );
	char* v = file_line( fp );
	int version = atoi( v );
	free( v );
	if ( fp != NULL ) {
		imv.loadState( fp, version );
		fclose( fp );
	}
}

void saveLine( const char* line ) {
	//vizState.saveLine( line );
}

CCubesView::CCubesView()
    : m_pEngine(NULL),
    m_pDiffuseColor(NULL),
    m_pTranslation(NULL),
    m_bCubeInScene(false)
{
}

CCubesView::~CCubesView()
{
    if (m_pEngine) m_pEngine->Release();
    if (m_pDiffuseColor) m_pDiffuseColor->Release();
    if (m_pTranslation) m_pTranslation->Release();
}

/////////////////////////////////////////////////////////////////////////////
// CCubesView drawing

void CCubesView::OnDraw(CDC* pDC)
{
    CCubesDoc* pDoc = GetDocument();
    ASSERT_VALID(pDoc);
}

/////////////////////////////////////////////////////////////////////////////
// CCubesView diagnostics

#ifdef _DEBUG
void CCubesView::AssertValid() const
{
    CView::AssertValid();
}

void CCubesView::Dump(CDumpContext& dc) const
{
    CView::Dump(dc);
}

CCubesDoc* CCubesView::GetDocument() // non-debug version is inline
{
    ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CCubesDoc)));
    return (CCubesDoc*)m_pDocument;
}
#endif //_DEBUG





const char* noextension( const char* fileName ) {
	const char* scanner = fileName;
	static char result[100];
	int ridx = 0;
	while ( *scanner ) {
		if (*scanner == '.') {
			result[ridx] = 0;
		} else if (( *scanner == '/' ) || ( *scanner == '\\' )) {
			ridx = -1;
			result[0] = 0;
		} else {
			result[ridx] = *scanner;
		}
		ridx++;
		scanner++;
	}
	return( newstrcpy( &result[0] ));
}

void openVrmlFile( const char* fileName ) {
	timerMode = SuperFast;
	log_message( "Adding openvrml file command" );
	log_message( (char*)fileName );

	// Wipe out the basepath variable, so we use full path 
	// that we got out of the file selection dialog
	// (really should save old one and restore it)
	if ( lastLabel != NULL ) {
		log_message1( "Creating a setInfo based on last label value", lastLabel );
		char buf[200];
		sprintf( buf, "info %s", lastLabel );
		//cq.insertMode(); // put on special insert queue
		cq.addCommand( SetInfo, newstrcpy( buf ));
		//cq.appendMode();
	} else {
		log_message( "setInfo but no real value" );
		cq.addCommand( SetInfo, newstrcpy( "not set" ));
	}
	FILE* fp = fopen( fileName, "r" );
	if ( fp == NULL ) {
		bool haspath = false;
		const char* scanner = fileName;
		while ( *scanner ) {
			if (*scanner == '\\') {
				haspath = true;
				break;
			} else if ( *scanner == '/' ) {
				haspath = true;
				break;
			} else {
				scanner++;
			}
		}
		// this first case allows "load xxx.png" to occur in config text script files...
		if ( !haspath ) {
			char buf[100];
			sprintf( buf, "C:/mindtel/render/vrml/%s", fileName );
			fp = fopen( buf, "r" );
		}
	}
	if ( fp == NULL ) {
		return;
	}
	char* line;
	//cq.insertMode();
	cq.suspend();
	cq.addCommand( StartData, noextension( fileName ));
	while (( line = file_line( fp )) != NULL ) {
		if ( strcmp( line, "addtexture" ) == 0 ) {
			if ( imageFile != NULL ) {
				char buf[200];
				sprintf( buf, "texture ImageTexture { url \"%s\" }", imageFile );
				cq.addCommand( Data, buf );
			}
		} else {
			cq.addCommand( Data, line );
		}
	}

	fclose( fp );
	cq.addCommand( EndData, NULL );
	cq.resume();
	//cq.appendMode();
}

char* negate( char* triplet ) {
	static char resultbuf[100];
	int resultidx = 0;
	if ( isdigit( *triplet ) || (*triplet == '.')) {
		resultbuf[resultidx++] = '-';
	} else if ( *triplet == '-' ) {
		triplet++;
	}
	while ( isdigit(*triplet) || (*triplet == '.')) {
		resultbuf[resultidx++] = *triplet++;
	}
	while ( !isdigit(*triplet) && (*triplet != '.') && (*triplet != '-')) {
		resultbuf[resultidx++] = *triplet++;
	}
	if ( isdigit( *triplet ) || (*triplet == '.')) {
		resultbuf[resultidx++] = '-';
	} else if ( *triplet == '-' ) {
		triplet++;
	}
	while ( isdigit(*triplet) || (*triplet == '.')) {
		resultbuf[resultidx++] = *triplet++;
	}
	while ( !isdigit(*triplet) && (*triplet != '.') && (*triplet != '-')) {
		resultbuf[resultidx++] = *triplet++;
	}
	if ( isdigit( *triplet ) || (*triplet == '.')) {
		resultbuf[resultidx++] = '-';
	} else if ( *triplet == '-' ) {
		triplet++;
	}
	while ( isdigit(*triplet) || (*triplet == '.')) {
		resultbuf[resultidx++] = *triplet++;
	}
	return( &resultbuf[0] );
}

static char* sphereSyntaxHeader = "Transform { children [ DirectionalLight { direction 0 -1 0 ambientIntensity 1 }";
static char* sphereSyntaxUse = "Transform { translation %s children [ Shape { appearance Appearance { material USE %s } geometry Sphere { radius %s }}]} Shape { appearance Appearance { material USE %s } geometry IndexedLineSet { coord Coordinate { point [ 0 0 0 %s ] } coordIndex [ 0 1 ] }}";
static char* sphereSyntaxDef = "Transform { translation %s children [ Shape { appearance Appearance { material DEF %s Material { diffuseColor %s emissiveColor %s }} geometry Sphere { radius %s }}]} Shape { appearance Appearance { material USE %s } geometry IndexedLineSet { coord Coordinate { point [ 0 0 0 %s ] } coordIndex [ 0 1 ] }}";
static char* sphereSyntaxFooter = "]}";

char* nullspace( char* scanner, int spacecount ) {
	int counter = 0;
	while ( *scanner ) {
		if ( *scanner == ' ' ) {
			counter++;
			if ( counter == spacecount ) {
				*scanner = 0;
				return( scanner + 1 );
			}
		}
		scanner++;
	}
	return( NULL );
}


class ObjectProcessor {
public:
	ObjectProcessor( const char* objectName );
protected:
	const char* objectName;
};

class SphereProcessor : public ObjectProcessor {
public:
	SphereProcessor( const char* objectName );
	void processLine( char* line );
	void generate();
private:
//	const char* objectName;
	int whichLine;
	int whichDef;
	char lastcolor[50];
	char materialDefName[50];
};

class MeshProcessor : public ObjectProcessor {
public:
	MeshProcessor( const char* objectName );
	void processLine( char* line );
	void generate();
private:
//	const char* objectName;
	int rows;
	int columns;
	char* points[1000];
	int pointidx;
};

void openImageFile( const char* fileName ) {
	const char* scanner = fileName;
	bool haspath = false;
	while ( *scanner ) {
		if (*scanner == '\\') {
			haspath = true;
			break;
		} else if ( *scanner == '/' ) {
			haspath = true;
			break;
		} else {
			scanner++;
		}
	}
	// this first case allows "load xxx.png" to occur in config text script files...
	if ( !haspath ) {
		char buf[100];
		sprintf( buf, "C:/mindtel/render/textures/%s", fileName );
		imageFile = newstrcpy( buf );
	} else {
		imageFile = newstrcpy( fileName );
	}
	log_message1( "image file: ", imageFile );
}


SphereProcessor::SphereProcessor( const char* o ) : ObjectProcessor( o ) {
	whichLine = 0;
	whichDef = 0;
	lastcolor[0] = 0;
	sprintf( materialDefName, "Mat_%s", o );
}

void SphereProcessor::processLine( char* line ) {
	char* translation = line;
	char* scanner = line;
	scanner = nullspace( scanner, 3 );
	if ( scanner == NULL ) {
		return;
	} else {
		//log_message1( "translation", translation );
	}
	char* radius = scanner;
	scanner = nullspace( scanner, 1 );
	if ( scanner == NULL ) {
		return;
	} else {
		//log_message1( "radius", radius );
	}
	whichLine++;
	char* color = scanner;
	char buf[1000];
	if ( whichLine == 1 ) {
		//log_message( "FIRST LINE" );
		sprintf( buf, sphereSyntaxHeader );
		//log_message( buf );
		cq.addCommand( Data, newstrcpy( buf ));
	}
	if ( strcmp( color, lastcolor ) == 0 ) {
		sprintf( buf, sphereSyntaxUse, translation, materialDefName, radius, materialDefName, translation );
	} else {
		whichDef++;
		strcpy( lastcolor, color );
		if ( whichDef == 1 ) {
			sprintf( materialDefName, "Mat_%s", objectName );
		} else {
			sprintf( materialDefName, "Mat_%s_%d", objectName, whichDef );
		}
		sprintf( buf, sphereSyntaxDef, translation, materialDefName, color, color, radius, materialDefName, translation);
	}
	//log_message( buf );
	cq.addCommand( Data, newstrcpy( buf ));
}

ObjectProcessor::ObjectProcessor( const char* o ) {
	objectName = o;
}

MeshProcessor::MeshProcessor( const char* o ) : ObjectProcessor( o ) {
//	objectName = o;
	rows = 0;
	columns = 0;
	pointidx = 0;
}

char* meshTextureSyntax[] = {
	"DirectionalLight {  direction 0 -1 0     ambientIntensity .7 } Shape { appearance Appearance { material DEF Mat_%s Material { diffuseColor .4 .8 .5 }         texture ImageTexture { url [ \"%s\" ] }}",
	"geometry IndexedFaceSet { coord Coordinate { point [",
	"] } coordIndex [",
	"] texCoord TextureCoordinate { point [ ",
	"]}}}"
};
char* meshSyntax[] = {
	"Shape { appearance Appearance { material DEF Mat_%s Material { emissiveColor 1 1 1 }}",
	"geometry IndexedLineSet { coord Coordinate { point [",
	"] } coordIndex [",
	"] }}"
};

void MeshProcessor::generate() {
	log_message1( "MeshProcessor::generate", objectName );
	char buf[1000];
	if ( imageFile == NULL ) {
		sprintf( buf, meshSyntax[0], objectName );
	} else {
		sprintf( buf, meshTextureSyntax[0], objectName, imageFile );
	}
	//log_message( buf );
	cq.addCommand( Data, newstrcpy( buf ));
	if ( imageFile == NULL ) {
		cq.addCommand( Data, meshSyntax[1] );
	} else {
		cq.addCommand( Data, meshTextureSyntax[1] );
		//log_message( meshTextureSyntax[1] );
	}
	int i = 0;
	int j = 0;
	for ( i = 0; i < pointidx; i++ ) {
		cq.addCommand( Data, points[i] );
		//log_message( points[i] );
	}
	log_message( meshSyntax[2] );
	cq.addCommand( Data, meshSyntax[2] );
	if ( imageFile == NULL ) {
		for ( i = 0; i < rows; i++ ) {
			buf[0] = 0;
			for ( j = 0; j < columns; j++ ) {
				char tbuf[20];
				sprintf( tbuf, "%d ", i*columns + j );
				strcat( buf, tbuf );
			}
			strcat( buf, "-1" );
			//log_message1( "mesh buf idx", buf );
			cq.addCommand( Data, newstrcpy( buf ));
		}
		for ( i = 0; i < columns; i++ ) {
			buf[0] = 0;
			for ( j = 0; j < rows; j++ ) {
				char tbuf[20];
				sprintf( tbuf, "%d ", j*columns + i );
				strcat( buf, tbuf );
			}
			strcat( buf, "-1" );
			//log_message1( "mesh buf idx", buf );
			cq.addCommand( Data, newstrcpy( buf ));
		}
		cq.addCommand( Data, meshSyntax[3] );
	} else {
		for ( i = 0; i < rows - 1; i++ ) {
			for ( j = columns; j > 0; j-- ) {
				char tbuf[40];
				sprintf( tbuf, "%d %d %d %d -1", i*columns + j - 2, i*columns + j - 1, (i+1)*columns+j-1, (i+1)*columns+j-2 );
				cq.addCommand( Data, newstrcpy( tbuf ));
				//log_message( tbuf );
			}
		}
		//log_message( meshTextureSyntax[3] );
		cq.addCommand( Data, meshTextureSyntax[3] );
		//log_message( "And now I am beginning loop to add texture coordinates" );
		for ( i = 0; i < rows; i++ ) {
			//log_messagei1( "Row", i );
			for ( j = columns - 1; j >= 0; j-- ) {
				char tbuf[40];
				sprintf( tbuf, "%f %f", (float)j/(float)(columns-1), 1.0 - (float)i/(float)(rows-1) );
				cq.addCommand( Data, newstrcpy( tbuf ));
				//log_message( tbuf );
			}
		}

		//log_message( meshTextureSyntax[4] );
		cq.addCommand( Data, meshTextureSyntax[4] );
		//log_message( "And that was it!" );
	}

}

	
void MeshProcessor::processLine( char* line ) {
	if (( rows == 0 ) || ( columns == 0 )) {
		line = nullspace( line, 1 );
		char* rowStr = line;
		if ( rowStr == NULL ) {
			return;
		}
		char* colStr = nullspace( line, 1 );
		if ( colStr == NULL ) {
			return;
		}
		rows = atoi( rowStr );
		columns = atoi( colStr );
		//log_messagei1( "rows: ", rows );
		//log_messagei1( "cols: ", columns );
	} else {
		points[ pointidx++ ] = line;
	}
}

void openSphereFile( const char* fileName ) {
	//log_message( "Opening Sphere file" );
	//log_message( fileName );

	// Wipe out the basepath variable, so we use full path 
	// that we got out of the file selection dialog
	// (really should save old one and restore it)
	//cq.addCommand( SetInfo, "basepath" );
	//cq.addCommand( Inline, (char*)fileName );
	//cq.suspend();
	cq.insertMode();
	cq.addCommand( SetInfo, fileName );
	const char* noextensionName = noextension( fileName );
	//log_message1( "sphere StartData param", noextensionName );
	cq.addCommand( StartData, noextensionName );
	FILE* fp = fopen( fileName, "r" );
	char* line;
	SphereProcessor sp( noextensionName );
	transformName = (char *)noextensionName;
	while (( line = file_line(fp) ) != NULL ) {
		//log_message( line );
		sp.processLine( line );
	}
	cq.addCommand( Data, newstrcpy( sphereSyntaxFooter ));
	cq.addCommand( EndData, NULL );
	log_message( "DONE WITH PUTTING SPHERE IN COMMAND QUEUE" );
	cq.appendMode();
	//cq.resume();
}

void openMeshFile( const char* fileName ) {
	//log_message( "Opening Mesh file" );
	//log_message( fileName );

	// Wipe out the basepath variable, so we use full path 
	// that we got out of the file selection dialog
	// (really should save old one and restore it)
	//cq.addCommand( SetInfo, "basepath" );
	//cq.addCommand( Inline, (char*)fileName );
	//cq.suspend();
	cq.insertMode();
	cq.addCommand( SetInfo, fileName );
	const char* noextensionName = noextension( fileName );
	//log_message1( "sphere StartData param", noextensionName );
	cq.addCommand( StartData, noextensionName );
	FILE* fp = fopen( fileName, "r" );
	char* line;
	MeshProcessor mp( noextensionName );
	while (( line = file_line(fp) ) != NULL ) {
		///log_message( line );
		mp.processLine( line );
	}
	fclose( fp );
	mp.generate();
	cq.addCommand( EndData, NULL );
	log_message( "DONE WITH PUTTING MESH IN COMMAND QUEUE" );
	cq.appendMode();
	//cq.resume();
}


void openCommandFile( const char* fileName ) {
	log_message1( "OPEN FILE:", fileName );
	if ( strlen( fileName ) > 4 ) {
		//log_message( "got step A" );
		if ( strncmp( fileName + strlen( fileName ) - 4, ".wrl", 4 ) == 0 ) {
			openVrmlFile( fileName );
			log_message( "WRL FILE LINES HAVE BEEN QUEUED" );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 7, ".sphere", 7 ) == 0 ) {
			log_message( "Opening sphere file" );
			openSphereFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 6, ".state", 6 ) == 0 ) {
			log_message( "Opening state file" );
			openStateFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 5, ".mesh", 5 ) == 0 ) {
			log_message( "Opening mesh file" );
			openMeshFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 4, ".jpg", 4 ) == 0 ) {
			log_message( "Opening jpg file" );
			openImageFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 4, ".png", 4 ) == 0 ) {
			log_message( "Opening png file" );
			openImageFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 4, ".JPG", 4 ) == 0 ) {
			log_message( "Opening jpg file" );
			openImageFile( fileName );
			return;
		} else if ( strncmp( fileName + strlen( fileName ) - 4, ".PNG", 4 ) == 0 ) {
			log_message( "Opening png file" );
			openImageFile( fileName );
			return;
		}
	}
	if (( lastFileName != NULL ) && ( strcmp( lastFileName, fileName ) == 0 )) {
		log_message1( "NOT OPENING, FILE ALREADY OPEN", lastFileName );
		return;
	}
	if ( lastFileName != NULL ) {
		free( lastFileName );
	}
	lastFileName = newstrcpy( fileName );
	//cq.insertMode();
	cq.load( fileName );
	//cq.appendMode();
	//log_message( "ALL LINES IN FILE HAVE BEEN PUT IN QUEUE" );
}


/**
 *  This is the default thread for accepting connections from neattools.
 *  The number of threads accepting connections is currently limited to 2
 *  (this and the following one).  This number needs to be configurable
 *  at startup time.
 */
UINT MyThreadProc( LPVOID pParam )
{
	int counter = 0;
	//log_message( "Creating worker thread");
	cq.load( "config.template" );
	//cq.load( "config.demo" );
	SocketCommand* sc = new SocketCommand(3502);
	while ( 1 ) {
		//log_message( "getting a command");
		char* command = sc->GetCommand();
		if ( strlen( command ) == 0 ) {
			Sleep( 100 );
		} else {
			cq.queue( command );
//			log_message( command );
		}
	}

    return 0;   // thread completed successfully
}

UINT MyThreadProc2( LPVOID pParam )
{
	int counter = 0;
	log_message( "2Creating worker thread");
	SocketCommand* sc = new SocketCommand(3503);
	while ( 1 ) {
		//log_message( "2getting a command");
		char* command = sc->GetCommand();
		if ( strlen( command ) == 0 ) {
			Sleep( 1000 );
		} else {
			cq2.queue( command );
		}
		//log_message( command );
	}
	log_message( "2thread is over");

    return 0;   // thread completed successfully
}

/////////////////////////////////////////////////////////////////////////////
// CCubesView message handlers


BOOL CCubesView::Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext) 
{
    // TODO: Add your specialized code here and/or call the base class
    if ( ! CWnd::Create(lpszClassName, lpszWindowName, dwStyle, rect, pParentWnd, nID, pContext) )
        return FALSE;

    if ( ! m_wndCortona.Create ( "Cortona VRML Control Window", dwStyle, rect, this, ID_CORTONA_CONTROL ) )
        return FALSE;

    // Turn on the headlight (some scenes without self illuminating materials
    // can be invisible when headlight is turned off)
    m_wndCortona.SetHeadLight ( TRUE );

    // Show navigation bars (to enable user to change Cortona control parameters 
    // such as navigation type, current viewpoint and so on visually)
    m_wndCortona.SetNavigationBar ( 1 );

    // Whenever any Cortona control parameter has been changed - call Refresh
    m_wndCortona.Refresh ();

    // It is a must call if you want to change scene anywhere
    m_wndCortona.Edit ();

	SetEngine();
 
	// one thread for collecting commands, timer for running them
	// This number of threads needs to be configurable, and the process needs to take a 
	// command queue as a parameter...
	AfxBeginThread(MyThreadProc, NULL);
	AfxBeginThread(MyThreadProc2, NULL);
	SetTimer(777, 1, NULL );

    return TRUE;
}

void CCubesView::SetEngine() {
    IDispatch* pdisp = m_wndCortona.GetEngine();

    if ( !pdisp || FAILED( pdisp->QueryInterface( IID_IEngine, (void**)&m_pEngine ))) {
        AfxMessageBox("Can not get Cortona Engine");
    } else {
		// The following instructs engine to immidiately apply
		// any changes in VRML nodes' fields values
		m_pEngine -> put_AutoRefresh ( TRUE );
	}
}

void CCubesView::OnSize(UINT nType, int cx, int cy) 
{
    CView::OnSize(nType, cx, cy);

    // Resize control window to fill all CView area
    if ( cx > 0 && cy > 0 )
        m_wndCortona.SetWindowPos ( NULL, 0, 0, cx, cy, SWP_NOZORDER | SWP_NOMOVE );
}

BOOL CCubesView::OnEraseBkgnd(CDC* pDC)
{
	return TRUE;
}

bool CCubesView::CreateVrmlFromString( /*in*/ OLECHAR* szVrml, /*out*/ INodeObject** ppRoot )
{
    if ( !m_pEngine || !ppRoot )
        return false;

    BSTR bstrVrmlSyntax = SysAllocString( szVrml );
    HRESULT hr = m_pEngine -> CreateNodeFromString ( bstrVrmlSyntax, ppRoot );
    SysFreeString( bstrVrmlSyntax );

    return SUCCEEDED( hr );
}

bool CCubesView::AddToScene( INodeObject* pSubsceneRoot )
{
	//log_message( "ADD TO SCENE" );
    if ( !m_pEngine || !pSubsceneRoot )
        return false;

    // We get root nodes collection
	IMFNodeObject* currentGroup = imv.getGroup();
	if ( currentGroup != NULL ) {
		//log_message( "NON-NULL CURRENT GROUP" );
		VARIANT before;
		before.vt = VT_ERROR;
		before.scode = DISP_E_PARAMNOTFOUND; 
		HRESULT hr = currentGroup -> Add ( pSubsceneRoot, &before );
		return SUCCEEDED( hr );
	} else {
		//log_message( "NULL CURRENT GROUP" );
		INodesCollection* pRoots = NULL;
		if ( pRoots == NULL ) {
			if ( FAILED( m_pEngine -> get_RootNodes ( &pRoots ) ) )
				return false;
		}

		// NOTE: you can pass the next VARIANT if an operand in method is optional
		VARIANT varDef;
		VariantInit( &varDef );
		varDef.vt = VT_ERROR;
		varDef.lVal = 0;

		// And add the given node to the root nodes collection
		HRESULT hr = pRoots->Add( pSubsceneRoot, &varDef );
		pRoots->Release ();
    
		return SUCCEEDED( hr );
	}
}


bool CCubesView::GetChildrenField( INodeObject* pSubsceneRoot, IMFNodeObject** ppChildren )
{
    if ( !m_pEngine || !ppChildren )
        return false;

	IFieldsCollection* pTransformFields;

    HRESULT hr = pSubsceneRoot -> get_Fields ( &pTransformFields );

    if ( SUCCEEDED ( hr ) ) {

        // Getting translation field by name
        IFieldObject* pTranslationField;

        VARIANT varName;
        VariantInit( &varName );
        varName.vt = VT_BSTR;
        varName.bstrVal = SysAllocString( L"children" );

        hr = pTransformFields -> get_Item ( &varName,  &pTranslationField );

        pTransformFields -> Release();
        VariantClear( &varName );

        if ( SUCCEEDED ( hr ) ) {

            // Get ISFVec3fObject from IFieldObject
            hr = pTranslationField->QueryInterface(IID_IMFNodeObject, (void**)ppChildren);
            pTranslationField -> Release();
        }
    }
    return SUCCEEDED( hr );
}

bool CCubesView::GetRotationField( INodeObject* pSubsceneRoot, ISFRotationObject** ppTranslation )
{
    if ( !m_pEngine || !ppTranslation )
        return false;

	IFieldsCollection* pTransformFields;

    HRESULT hr = pSubsceneRoot -> get_Fields ( &pTransformFields );

    if ( SUCCEEDED ( hr ) ) {

        // Getting translation field by name
        IFieldObject* pTranslationField;

        VARIANT varName;
        VariantInit( &varName );
        varName.vt = VT_BSTR;
        varName.bstrVal = SysAllocString( L"rotation" );

        hr = pTransformFields -> get_Item ( &varName,  &pTranslationField );

        pTransformFields -> Release();
        VariantClear( &varName );

        if ( SUCCEEDED ( hr ) ) {

            // Get ISFVec3fObject from IFieldObject
            hr = pTranslationField->QueryInterface(IID_ISFRotationObject, (void**)ppTranslation);
            pTranslationField -> Release();
        }
    }
    return SUCCEEDED( hr );
}


bool CCubesView::GetTranslationField( INodeObject* pSubsceneRoot, ISFVec3fObject** ppTranslation )
{
    if ( !m_pEngine || !ppTranslation )
        return false;

	IFieldsCollection* pTransformFields;

    HRESULT hr = pSubsceneRoot -> get_Fields ( &pTransformFields );

    if ( SUCCEEDED ( hr ) ) {

        // Getting translation field by name
        IFieldObject* pTranslationField;

        VARIANT varName;
        VariantInit( &varName );
        varName.vt = VT_BSTR;
        varName.bstrVal = SysAllocString( L"translation" );

        hr = pTransformFields -> get_Item ( &varName,  &pTranslationField );

        pTransformFields -> Release();
        VariantClear( &varName );

        if ( SUCCEEDED ( hr ) ) {

            // Get ISFVec3fObject from IFieldObject
            hr = pTranslationField->QueryInterface(IID_ISFVec3fObject, (void**)ppTranslation);
            pTranslationField -> Release();
        }
    }
    return SUCCEEDED( hr );
}


bool CCubesView::GetScaleField( INodeObject* pSubsceneRoot, ISFVec3fObject** ppTranslation )
{
    if ( !m_pEngine || !ppTranslation )
        return false;

	IFieldsCollection* pTransformFields;


    // Here we try to get Xform_<name> instead of using transform added
 /*   INodesCollection* pNodes;
    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;
    INodeObject* pXform;
    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Xform_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();
    HRESULT hr = pNodes -> get_Item ( &varName, &pXform );
    pNodes -> Release();
    VariantClear( &varName );
    if ( SUCCEEDED ( hr ))
        pSubsceneRoot = pXform;*/


	// continue as before
    HRESULT hr = pSubsceneRoot -> get_Fields ( &pTransformFields );

    if ( SUCCEEDED ( hr ) ) {

        // Getting translation field by name
        IFieldObject* pTranslationField;

        VARIANT varName;
        VariantInit( &varName );
        varName.vt = VT_BSTR;
        varName.bstrVal = SysAllocString( L"scale" );

        hr = pTransformFields -> get_Item ( &varName,  &pTranslationField );

        pTransformFields -> Release();
        VariantClear( &varName );

        if ( SUCCEEDED ( hr ) ) {

            // Get ISFVec3fObject from IFieldObject
            hr = pTranslationField->QueryInterface(IID_ISFVec3fObject, (void**)ppTranslation);
            pTranslationField -> Release();
        }
    }
    return SUCCEEDED( hr );
}


bool CCubesView::GetScriptValue( INodeObject* pSubsceneRoot, ISFInt32Object** ppIn1, ISFInt32Object** ppIn2, ISFInt32Object** ppIn3, ISFInt32Object** ppIn4, char* name )
{
    if ( !m_pEngine || !ppIn1 )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Script_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, &pMaterial );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ))
        return false;

    // get all node fields
    IFieldsCollection* pScriptFields;

    hr = pMaterial -> get_Fields ( &pScriptFields );
    pMaterial -> Release();

    if ( FAILED ( hr ) )
        return false;

    // get field by name
    IFieldObject* pScriptField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"in1" );
    hr = pScriptFields -> get_Item ( &varName,  &pScriptField );
    VariantClear( &varName );
    if ( FAILED ( hr ) ) {
		pScriptFields -> Release();
        return false;
	}
    hr = pScriptField -> QueryInterface ( IID_ISFInt32Object, (void**)ppIn1 );
    pScriptField -> Release();

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"in2" );
    hr = pScriptFields -> get_Item ( &varName,  &pScriptField );
    VariantClear( &varName );
    if ( !FAILED ( hr ) ) {
	    hr = pScriptField -> QueryInterface ( IID_ISFInt32Object, (void**)ppIn2 );
		pScriptField -> Release();
	}

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"in3" );
    hr = pScriptFields -> get_Item ( &varName,  &pScriptField );
    VariantClear( &varName );
    if ( !FAILED ( hr ) ) {
	    hr = pScriptField -> QueryInterface ( IID_ISFInt32Object, (void**)ppIn3 );
		pScriptField -> Release();
	}

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"in4" );
    hr = pScriptFields -> get_Item ( &varName,  &pScriptField );
    pScriptFields -> Release();
    VariantClear( &varName );
    if ( !FAILED ( hr ) ) {
	    hr = pScriptField -> QueryInterface ( IID_ISFInt32Object, (void**)ppIn4 );
		pScriptField -> Release();
	}

	log_message( "I SUCCESSFULLY GOT A SCRIPT INPUT!" );
	if ( *ppIn1 != NULL ) {
		log_message( "and I mean I really got it!" );
	}
    return SUCCEEDED( hr );
}


/*
 * ======== CCUbesView::GetEmbeddedTransform ========
 *
 *  Purpose:
 *    When a visual object gets added to renderer, it is embedded in a Transform node
 *    from the "config.template" file.  However, some operations, such as rotation and
 *    scaling depend on a "center" field, which means they have to refer to the embedded
 *    Transform and not the top-level Transform node.
 *
 *    This method finds that embedded Transform by searching for a DEF Xform_<name>
 *
 *  Input Parameters:
 *    pSubsceneRoot -- top level Transform, where search begins
 *    transformName -- name of the Transform DEF
 *
 *  Output Parameters:
 *    embeddedTransform -- if found, this points to result.  Otherwise it is unchanged.
 */

bool CCubesView::GetEmbeddedTransform( INodeObject* pSubsceneRoot, INodeObject** embeddedTransform, char* name )
{
    if ( !m_pEngine || !embeddedTransform )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Xform_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, embeddedTransform );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ))
        return false;


    return SUCCEEDED( hr );
}


bool CCubesView::GetDiffuseColorField( INodeObject* pSubsceneRoot, ISFColorObject** ppDiffuseColor, char* name )
{
    if ( !m_pEngine || !ppDiffuseColor )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Mat_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes->get_Item ( &varName, &pMaterial );

    pNodes->Release();
    VariantClear( &varName );

    if ( FAILED ( hr )) {
		//log_message1( "Did not find DEF Mat_", name );
        return false;
	}

    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pMaterial -> get_Fields ( &pMaterialFields );
    pMaterial -> Release();

    if ( FAILED ( hr ) )
        return false;

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"diffuseColor" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ) )
        return false;

    hr = pMaterialField -> QueryInterface ( IID_ISFColorObject, (void**)ppDiffuseColor );

    pMaterialField -> Release();

    return SUCCEEDED( hr );
}

bool CCubesView::GetMarkerLocation( INodeObject* pSubsceneRoot, ISFVec3fObject** ppMarker, char* name )
{
	//log_message1( "Attempting to get Marker", name );
    if ( !m_pEngine || !ppMarker ) {
		//log_message( "false return #1" );
        return false;
	}

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED( m_pEngine->get_Nodes( &pNodes ))) {
		//log_message( "false return #2" );
        return false;
	}

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes->get_Item( &varName, &pMaterial );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr )) {
		//log_message( "false return #3" );
        return false;
	}

    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pMaterial->get_Fields ( &pMaterialFields );
    pMaterial->Release();

    if ( FAILED( hr )) {
		//log_message( "false return #4" );
        return false;
	}

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"markLocation" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr )) {
		//log_message( "false return #5" );
        return false;
	}

    hr = pMaterialField->QueryInterface( IID_ISFVec3fObject, (void**)ppMarker );

    pMaterialField->Release();

	/*if ( SUCCEEDED( hr )) {
		log_message("found marker" );
	} else {
		log_message("did not find marker" );
	}*/
    return SUCCEEDED( hr );
}



bool CCubesView::GetTransparencyField( INodeObject* pSubsceneRoot, ISFFloatObject** ppTransparency, char* name )
{
	//log_message1( "GetTransparencyField for ", name );
    if ( !m_pEngine || !ppTransparency )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Mat_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, &pMaterial );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ))
        return false;

    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pMaterial -> get_Fields ( &pMaterialFields );
    pMaterial -> Release();

    if ( FAILED ( hr ) ) {
		log_message( "did not get material fields" );
        return false;
	}

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"transparency" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ) ) {
		log_message( "did not get transparency" );
        return false;
	}

    hr = pMaterialField -> QueryInterface ( IID_ISFFloatObject, (void**)ppTransparency );

    pMaterialField -> Release();

	//log_message( "SUCCESS" );
    return SUCCEEDED( hr );
}


bool CCubesView::GetLabelField( INodeObject* pSubsceneRoot, ISFStringObject** ppString )
{
	//log_message1( "GetTransparencyField for ", name );
    if ( !m_pEngine || !ppString )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "info_script" );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, &pMaterial );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ))
        return false;

    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pMaterial -> get_Fields ( &pMaterialFields );
    pMaterial -> Release();

    if ( FAILED ( hr ) ) {
		log_message( "did not get material fields" );
        return false;
	}

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"inString" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ) ) {
		log_message( "did not get transparency" );
        return false;
	}

    hr = pMaterialField -> QueryInterface ( IID_ISFStringObject, (void**)ppString );

    pMaterialField -> Release();

	//log_message( "SUCCESS" );
    return SUCCEEDED( hr );
}


/**
 * ======== CCubesView::GetEmissiveColorField ========
 *
 *  Purpose:
 *    Get the ISFColorObject associated with an "emissiveColor" field.
 *
 *  Input Parameters:
 *    pSubsceneRoot -- seems to be unused
 *
 *  Output Parameters:
 *    ppEmissiveColor -- pointer to the resulting ISFColorObject if found
 *
 *  Return Value:
 *    true if it is found, otherwise false
 */
bool CCubesView::GetEmissiveColorField( INodeObject* pSubsceneRoot, ISFColorObject** ppEmissiveColor, char* name )
{
    if ( !m_pEngine || !ppEmissiveColor )
        return false;

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) )
        return false;

    // here we get node by its' DEF name
    INodeObject* pMaterial;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	CString m = CString( "Mat_" ) + CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, &pMaterial );

    pNodes -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ))
        return false;

    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pMaterial -> get_Fields ( &pMaterialFields );
    pMaterial -> Release();

    if ( FAILED ( hr ) )
        return false;

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"emissiveColor" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ) )
        return false;

    hr = pMaterialField -> QueryInterface ( IID_ISFColorObject, (void**)ppEmissiveColor );

    pMaterialField -> Release();

    return SUCCEEDED( hr );
}


bool CCubesView::TestNode( char* name, ISFFloatObject** ppTransparency )
{
	log_message( "CCubesView::TestNode entered" );
    if ( !m_pEngine ) {
		log_message( "CCubesView::TestNode false returned #1" );
        return false;
	}

    // the next gives us the whole scene node collection (on any level)
    INodesCollection* pNodes;

    if ( FAILED ( m_pEngine -> get_Nodes ( &pNodes ) ) ) {
		log_message( "CCubesView::TestNode false returned #2" );
        return false;
	}

    // here we get node by its' DEF name
    INodeObject* pNode;

    VARIANT varName;
    VariantInit( &varName );
    varName.vt = VT_BSTR;
	log_message1( "CCubesView::TestNode name in is ", name );
	CString m = CString( name );
    varName.bstrVal = m.AllocSysString();

    HRESULT hr = pNodes -> get_Item ( &varName, &pNode );

    pNodes -> Release();
    VariantClear( &varName );

	INodeObject* pTerrainMaterial = NULL;
    if ( FAILED ( hr )) {
		log_message( "CCubesView::TestNode false returned #3" );
        return false;
	} else {
		log_message( "CCubesView::TestNode true returned" );
		CString t2( "Terrain_Material" );
		if ( FAILED( pNode->get_Nodes( &pNodes ))) {
			log_message( "CCubesView::TestNode Failed trying to get Terrain_Material" );
			return false;
		} else {
			VARIANT varName2;
			VariantInit( &varName2 );
			varName2.vt = VT_BSTR;
			varName2.bstrVal = t2.AllocSysString();
			HRESULT hr2 = pNodes->get_Item( &varName2, &pTerrainMaterial );
			if ( FAILED( hr2 )) {
				log_message( "CCubesView::TestNode Failed getting Terrain_Material from Inline Nodes" );
				return false;
			} else {
				log_message( "CCubesView::TestNode Hey, I got Terrain_Material from Inline Nodes" );
			}
		}
		//return true;
	}
	    // get all node fields
    IFieldsCollection* pMaterialFields;

    hr = pTerrainMaterial -> get_Fields ( &pMaterialFields );
    pTerrainMaterial -> Release();

    if ( FAILED ( hr ) ) {
		log_message( "CCubesView::TestNode did not get material fields" );
        return false;
	}

    // get field by name
    IFieldObject* pMaterialField;

    VariantInit( &varName );
    varName.vt = VT_BSTR;
    varName.bstrVal = SysAllocString( L"transparency" );

    hr = pMaterialFields -> get_Item ( &varName,  &pMaterialField );
    pMaterialFields -> Release();
    VariantClear( &varName );

    if ( FAILED ( hr ) ) {
		log_message( "CCubesView::TestNode did not get transparency" );
        return false;
	}

    hr = pMaterialField -> QueryInterface ( IID_ISFFloatObject, (void**)ppTransparency );

    pMaterialField -> Release();

	log_message( "CCubesView::TestNode SUCCESS" );
    return SUCCEEDED( hr );

}


class DEFgen {
public:
	DEFgen();
	char* generate();
private:
	int counter;
};

DEFgen generator;

DEFgen::DEFgen() {
	counter = 0;
}

char* DEFgen::generate() {
	counter++;
	char buf[100];
	sprintf( buf, "label_%d", counter );
	return( newstrcpy( buf ));
}




char* getObjectAt( int offset ) {
	return( imv.getObjectAt( offset ));
}

bool isVisible( int offset ) {
	return( imv.isVisible( offset ));
}

// Create transform and add it to the scene
void CCubesView::AddTransformString(OLECHAR* syntax, char* label, char* defName, char* externalName ) {
	//log_message( "ADD TRANFORM STRING" );
    INodeObject* pTransformNode = NULL;
    if ( CreateVrmlFromString (syntax, &pTransformNode ) )    {
		VisualObject* vo = new VisualObject( pTransformNode );
        if ( AddToScene ( pTransformNode ) ) {
			ISFInt32Object* script_in1 = NULL;
			ISFInt32Object* script_in2 = NULL;
			ISFInt32Object* script_in3 = NULL;
			ISFInt32Object* script_in4 = NULL;
			GetScriptValue( pTransformNode, &script_in1, &script_in2, &script_in3, &script_in4, transformName );
			vo->SetScriptInputs( script_in1, script_in2, script_in3, script_in4, NULL );
			INodeObject* embeddedTransform = NULL;
			GetEmbeddedTransform( pTransformNode, &embeddedTransform, transformName );
			vo->SetEmbeddedTransform( embeddedTransform );
			ISFVec3fObject* pScale = NULL;
			if ( embeddedTransform != NULL ) {
				GetScaleField( embeddedTransform, &pScale );
				vo->SetScale( pScale );
			} 
			if ( pScale == NULL ) {
				GetScaleField( pTransformNode, &pScale );
				vo->SetScale( pScale );
			}
			ISFRotationObject* pRotation = NULL;
			if ( GetRotationField( pTransformNode, &pRotation )) {
				vo->SetRotation( pRotation );
			}
			IMFNodeObject* pGroup = NULL;
			if ( GetChildrenField( pTransformNode, &pGroup )) {
				vo->SetGroup( pGroup );
			}
			ISFStringObject* pString = NULL;
			if ( GetLabelField( pTransformNode, &pString )) {
				vo->SetLabel( pString );
			}
			vo->setLabelString( label );
			ISFVec3fObject* pMarker = NULL;
			if ( GetMarkerLocation( pTransformNode, &pMarker, transformName )) {
				log_message1("We got the marker location for ", transformName );
				vo->SetMarkerLocation( pMarker );
			}

			//log_message1( "defName is ", defName );
			//log_message1( "externalName is ", externalName );
			imv.addObject( this, pTransformNode, vo, defName, externalName );
			m_bCubeInScene = true;
        } else {
			log_message( "ADD TO SCENE FAILED" );
		}
        pTransformNode -> Release ();
    } else {
		log_message( "CREATE VRML FROM STRING FAILED" );
	}
} 



/*
 * ======== CCubesView::loadInline ========
 *
 *  Purpose:
 *    Load a named file using a template.  Used specifically for "inline"
 *
 *    The "inlineSyntax" is hardcoded to be for a specific directory, however, 
 *    the "config.template" file can override the "inline" syntax.
 *
 *  Input Parameters:
 *    name -- the parameter to insert into the "inlineSyntax" to get what we need
 */

static char* inlineSyntax = "Transform { children [  Inline { url \"c:\\mindtel\\render\\terrain\\%s\" } ] }";
void CCubesView::loadInline( char* name ) {
	if ( stm.getTemplateChars( "inline" ) != NULL ) {
		inlineSyntax = stm.getTemplateChars( "inline" );
		/*if ( svars.getTemplateChars( "basepath" ) == NULL ) {
			log_message( "we have no basepath setting" );
		} else {
			log_message1( "basepath is set to", svars.getTemplateChars( "basepath" ));
		}
		log_message1( "Before merging in basepath", inlineSyntax );*/
		SceneTemplate* inlineTemplate = stm.getTemplate( "inline" );
		inlineSyntax = inlineTemplate->mergeVariable( "basepath", 
						svars.getTemplateChars("basepath"));
		//log_message1( "After merging in basepath", inlineSyntax );
	}
	char* buf = (char*)malloc( strlen( inlineSyntax ) + strlen( name ) + 100 );
	sprintf( buf, inlineSyntax, name );
	log_message( buf );
	CString s( buf );
	AddTransformString( s.AllocSysString(), "", "", "" );
	free( buf );
}


void CCubesView::OnAddCube() 
{
	//AddTransformString( szCubeSyntax );
    //INodeObject* pCubeTransform = NULL;

    // Create cube and add it to the scene
    /*if ( CreateVrmlFromString ( szCubeSyntax, &pCubeTransform ) )
    {
        if ( AddToScene ( pCubeTransform ) )
        {        
            // OK, our cube is on stage, now we'll get the fields for future use
            GetTranslationField( pCubeTransform, &m_pTranslation );

            // Now we've got translation field
            // and we want to get material's field diffuseColor
            GetDiffuseColorField( pCubeTransform, &m_pDiffuseColor );
			GetTransparencyField( pCubeTransform, &m_pTransparency );

            m_bCubeInScene = true;
        }
        pCubeTransform -> Release ();
    }*/
}


void CCubesView::OnRemoveCube() 
{
    // All we have to do is to clear root nodes collection,
    // it's also possible to use Remove method of collection
    // to remove node by name or by index

	INodesCollection* pRoots;
    HRESULT hr = m_pEngine -> get_RootNodes ( &pRoots );

    if ( SUCCEEDED ( hr ) ) {
        pRoots -> Clear ();
        pRoots -> Release ();
    }

    if ( m_pDiffuseColor )
    {
        m_pDiffuseColor -> Release ();
        m_pDiffuseColor = NULL;
    }

    if ( m_pTranslation )
    {
        m_pTranslation -> Release ();
        m_pTranslation = NULL;
    }

    m_bCubeInScene = false;
}


/**
 * ======== CCubesView::OnMoveUp ========
 *
 *  Purpose:
 *    Move the current object up one unit.
 */

void CCubesView::OnMoveUp() {
	imv.MoveUp();
}


/**
 * ======== CCubesView::OnMoveDown ========
 *
 *  Purpose:
 *    Move the current object down one unit.
 */

void CCubesView::OnMoveDown() {
	imv.MoveDown();
}


/**
 * ======== CCubesView::OnMoveRight ========
 *
 *  Purpose:
 *    Move the current object to the right one unit.
 */

void CCubesView::OnMoveRight() {
	imv.MoveRight();
}

void CCubesView::OnMoveLeft() {
	imv.MoveLeft();
}

void CCubesView::OnMoveForward() {
	imv.MoveIn();
}

void CCubesView::OnMoveBackward() {
	imv.MoveOut();
}

void CCubesView::OnTaller() {
	imv.Taller();
}
void CCubesView::OnShorter() {
	imv.Shorter();
}
void CCubesView::OnFatter() {
	imv.Fatter();
}
void CCubesView::OnFadeOut() {
	imv.FadeOut();
}
void CCubesView::OnFadeIn() {
	imv.FadeIn();
}
void CCubesView::OnCycle() {
	imv.Cycle();
	OnJiggle();
}
void CCubesView::OnThinner() {
	imv.Thinner();
}

/**
 * ======== CCubesView::presentLabel ========
 *
 *  Purpose:  
 *    Put a text label into the static Text3D node, and add it to the scene
 */
void CCubesView::presentLabel( char* name ) {
	static char* labelSyntax = "Transform { children [  Shape { appearance Appearance { material DEF Mat_%s Material {	diffuseColor .8 0.3 .30 transparency .3	} } geometry Text3D { string [ \"%s\" ] }}]}";
	if ( stm.getTemplateChars( "label" ) != NULL ) {
		labelSyntax = stm.getTemplateChars( "label" );
	}
	char* buf = (char*)malloc( strlen( labelSyntax ) + strlen( name ) + 100 );
	transformName = generator.generate();
	sprintf( buf, labelSyntax, transformName, name );
	log_message( buf );
	CString s( buf );
	AddTransformString( s.AllocSysString(), NULL, transformName, transformName );
	free( buf );
}


/**
 * ======== sendLabel ========
 *
 *  Purpose:
 *    Adds a text label command to the command queue.  Used by the dialog box
 *
 *  Input Parameters
 *    s -- the string to use as a text label
 */
void sendLabel( CString s ) {
	log_message1("got the label: ", s.GetBuffer(100));
	// was this, changing it to a function call...	cq.addCommand( Label, newstrcpy( s.GetBuffer(100) ));
	//functionManager.execute( &cq, s.GetBuffer(100));
	imv.SetLabel( s.GetBuffer(100) );
}

void CCubesView::OnChangeColor() 
{
	log_message( "attempting to change diffuse color" );
	changeColor( imv.getDiffuseColor(), imv.getEmissiveColor() );
}

void CCubesView::changeColor( ISFColorObject* m_pDiffuseColor, ISFColorObject* m_pEmissiveColor ) {
	if (( m_pDiffuseColor == NULL ) && ( m_pEmissiveColor == NULL )) {
		log_message( "color is NULL" );
		return;
	}
	ISFColorObject* changer = m_pDiffuseColor;
	if ( changer == NULL ) {
		changer = m_pEmissiveColor;
	}

    float r, g, b;

    if ( SUCCEEDED ( changer -> get_Red ( &r ) ) &&
         SUCCEEDED ( changer -> get_Green ( &g ) ) &&
         SUCCEEDED ( changer -> get_Blue ( &b ) ))
    {
		char xbuf[100];
		sprintf( xbuf, "r %f, g %f, b %f\n", r, g, b );
		log_message( xbuf );
        CColorDialog colorDialog ( RGB ( r * 255, g * 255, b * 255 ) );

        if ( colorDialog.DoModal () == IDOK )
        {
            COLORREF newColor = colorDialog.GetColor ();

            r = ( float ) GetRValue ( newColor ) / 255.f;
            g = ( float ) GetGValue ( newColor ) / 255.f;
            b = ( float ) GetBValue ( newColor ) / 255.f;

			sprintf( xbuf, "new r %f, g %f, b %f\n", r, g, b );	
			log_message( xbuf );

			if ( m_pDiffuseColor != NULL ) {
				m_pDiffuseColor -> put_Red ( r );
				m_pDiffuseColor -> put_Green ( g );
				m_pDiffuseColor -> put_Blue ( b );
			}
			/*if ( m_pEmissiveColor != NULL ) {
				m_pEmissiveColor -> put_Red ( r );
				m_pEmissiveColor -> put_Green ( g );
				m_pEmissiveColor -> put_Blue ( b );
			}*/
        }
    } else {
		log_message( "Failed getting color values" );
	}
}

// Handle toolbar commands and UI command updates
void CCubesView::OnUpdateAddCube(CCmdUI* pCmdUI) 
{
    pCmdUI -> Enable ( ! m_bCubeInScene );
}
 
void CCubesView::OnUpdateCubeInScene(CCmdUI* pCmdUI) 
{
    pCmdUI -> Enable ( m_bCubeInScene );
}

//////////////////////////////////////////////////////////////////////
int counter = 0;
char* redoTestNodeParam = NULL;
void CCubesView::OnTimer(UINT nIDEvent) 
{
	// TODO: Add your message handler code here and/or call default
    // control will load and render the given VRML file
	
	Command* command = cq.getNextCommand();
	CommandQueue* cqp = &cq;
	if ( command == NULL ) {
		command = cq2.getNextCommand();
		cqp = &cq2;
	}
	//if ( command != NULL ) {
	while ( command != NULL ) {
		//log_message1( "cmd: ", cq.toStr( command->getValue()) );
		if (( command->getValue() == StartData ) || ( command->getValue() == ModelSetCommand )) {
			bool modelSetActive = false;
			int modelSetCount = 0;
			//log_message( "checking for modelset" );
			if ( command->getValue() == ModelSetCommand ) {
				modelSetActive = true;
				modelSetCount = atoi( command->getParam2() );
			}
			//log_message( "got start data or modelset command" );
			transformName = command->takeParam();
			log_message1( "Creating: ", transformName );
			delete command;
			SceneTemplate* st = new SceneTemplate( "internal" );
			while ( command = cqp->unconditionalGetNextCommand() ) {
				if ( command->getValue() == Data ) {
					//log_message( "got data" );
					st->addLine( command->getParam() );
				} else if ( command->getValue() == EndData ) {
					//log_message( "got end data" );
					break;
				}
				delete command;
			}
			SceneTemplate* transformTemplate = stm.getTemplate( "transform" );
			char* charsToUse = st->getChars();
			if ( transformTemplate != NULL ) {
				charsToUse = transformTemplate->mergeChars( st->getChars() );
				//log_message( "USING MERGED DATA" );
				//log_message( charsToUse );
				transformTemplate->scanVariables();
				char* variableName = NULL;
				while (( variableName = transformTemplate->getNextVariable() ) != NULL ) {
					//log_message( "VARIABLE TO USE" );
					//log_message( variableName );
					charsToUse = transformTemplate->mergeVariable( variableName, 
						svars.getTemplateChars(variableName));
				}
				//log_message( "XXXXXXXXXXXUSING MERGED DATA" );
				//log_message( charsToUse );
			} else {
				log_message( "NO transform template" );
			}
			CString result( charsToUse );
			//log_message( "object is being added to renderer" );
			//log_message( result.GetBuffer(10000) );
			SceneVariable* sv = svars.getVariable( "info" );
			char* varvalue = NULL;
			if ( sv != NULL ) {
				varvalue = sv->getChars();
			}
			/** The model set adds the same object many times, and registers the set
			 *  with a ModelSetManager.  The ModelSetManager allows external programs
			 *  like neattools to treat the model set as a single model.
			 */
			if ( modelSetActive ) {
				char baseName[100];
				char xname[100];
				strcpy( baseName, transformName );
				//log_message( "model set is active" );
				for ( int i = 0; i < modelSetCount; i++ ) {
					sprintf( xname, "%s_%d", baseName, i );
					transformName = newstrcpy( &xname[0] );
					//log_message1( "transformName is ", transformName );
					AddTransformString( result.AllocSysString(), varvalue, baseName, transformName );
				}
				imv.RegisterModelSet( baseName, modelSetCount );
				//log_message( "loop is done" );
			} else {
				AddTransformString( result.AllocSysString(), varvalue, transformName, transformName );
			}
			delete st;
			//log_message( "Object added" );
		} else if ( command->getValue() == StartTemplate ) {
			//log_message( "got start template command" );
			transformName = command->takeParam();
			stm.addTemplate( transformName );
			delete command;
			while ( command = cqp->unconditionalGetNextCommand() ) {
				//log_message( "got another command" );
				if ( command->getValue() == TemplateData ) {
					//log_message( "it is template data" );
					stm.addLine( command->getParam());
				} else {
					break;
				}
				delete command;
			}
			//log_message( "Finished adding template" );
			//log_message1( "Template text is: ", stm.getTemplateChars( transformName ));
		} else if ( command->getValue() == StartFunction ) {
			char* functionName = command->takeParam();
			Function* f = new Function( functionName );
			delete command;
			while ( command = cqp->unconditionalGetNextCommand() ) {
				if ( command->getValue() == FunctionData ) {
					f->addLine( command->getParam() );
				} else {
					break;
				}
				delete command;
			}
			functionManager.addFunction( f );
		} else if ( command->getValue() == Load ) {   
			log_message1( "Loading", command->getParam() );
			openCommandFile( command->getParam() );
		} else if ( command->getValue() == ReSetEngine ) {
			SetEngine();
		} else if ( command->getValue() == Hide ) {
			imv.Hide();
		} else if ( command->getValue() == Show ) {
			imv.Show();
		} else if ( command->getValue() == Left ) {
			imv.MoveLeft();
		} else if ( command->getValue() == Right ) {
			imv.MoveRight();
		} else if ( command->getValue() == Up ) {
			imv.MoveUp();
		} else if ( command->getValue() == Down ) {
			imv.MoveDown();
		} else if ( command->getValue() == In ) {
			imv.MoveIn();
		} else if ( command->getValue() == Out ) {
			imv.MoveOut();
		} else if ( command->getValue() == Taller ) {
			imv.Taller();
		} else if ( command->getValue() == Shorter ) {
			imv.Shorter();
		} else if ( command->getValue() == Bigger ) {
			imv.Bigger();
		} else if ( command->getValue() == Smaller ) {
			imv.Smaller();
		} else if ( command->getValue() == Fatter ) {
			imv.Fatter();
		} else if ( command->getValue() == Thinner ) {
			imv.Thinner();
		} else if ( command->getValue() == Cycle ) {
			imv.Cycle();
			OnJiggle();
		} else if ( command->getValue() == SetMax ) {
			imv.SetMax( command->getParam() );
		} else if ( command->getValue() == FadeOut ) {
			if ( redoTestNodeParam != NULL ) {
				ISFFloatObject* pTransparency = NULL;
				TestNode( redoTestNodeParam, &pTransparency );
				imv.setTransparency( pTransparency );
				redoTestNodeParam = NULL;
			}
			imv.FadeOut();
		} else if ( command->getValue() == TestDEF ) {
			ISFFloatObject* pTransparency = NULL;
			TestNode( command->getParam(), &pTransparency );
			imv.setTransparency( pTransparency );
		} else if ( command->getValue() == FadeIn ) {
			imv.FadeIn();
		} else if ( command->getValue() == Inline ) {
			strcpy( transformName, "terrain" );  // for now only one terrain allowed
			loadInline( command->getParam() );
		} else if ( command->getValue() == SetGroup ) {
			imv.SetGroup();
		} else if ( command->getValue() == Debug ) {
			cq.debugOn();
			cq2.debugOn();
		} else if ( command->getValue() == Label ) {
			log_message1("execute time, got param: ", command->getParam() );
			presentLabel( command->getParam() );
		} else if ( command->getValue() == SetObject ) {
			imv.SetObject( command->getParam() );
			//log_message1("object set to", command->getParam() );
		} else if ( command->getValue() == SetVizProperty ) {
			imv.SetProperty( command->getParam() );
		} else if ( command->getValue() == Clear ) {
			OnRemoveCube();
			lastFileName = NULL;
			imv.Clear();
		} else if ( command->getValue() == SetVizValue ) {
			imv.SetValue( command->getParam() );
		} else if ( command->getValue() == SetInfo ) {
			//log_message1( "WE GOT SETINFO", command->getParam() );
			if ( strcmp( command->getParam(), "info" ) == 0 ) {
				lastLabel = newstrcpy( command->getParam2() );
			}
			svars.addVariable( command->getParam(), command->getParam2() );
			if ( strcmp( command->getParam(), "fade_units" ) == 0 ) {
				log_message1( "Setting fade_units to", command->getParam2() );
				imv.SetFadeUnits( command->getParam2() );
			} else if ( strcmp( command->getParam(), "scale_units" ) == 0 ) {
				log_message1( "Setting scale units to", command->getParam2() );
				imv.SetScaleUnits( command->getParam2() );
			} else if ( strcmp( command->getParam(), "xmin" ) == 0 ) {
				imv.SetXmin( command->getParam2() );
			} else if ( strcmp( command->getParam(), "xmax" ) == 0 ) {
				imv.SetXmax( command->getParam2() );
			} else if ( strcmp( command->getParam(), "ymin" ) == 0 ) {
				imv.SetYmin( command->getParam2() );
			} else if ( strcmp( command->getParam(), "ymax" ) == 0 ) {
				imv.SetYmax( command->getParam2() );
			} else if ( strcmp( command->getParam(), "zmin" ) == 0 ) {
				imv.SetZmin( command->getParam2() );
			} else if ( strcmp( command->getParam(), "zmax" ) == 0 ) {
				imv.SetZmax( command->getParam2() );
			}
			//char* test = svars.getTemplateChars("info");
			//if ( test == NULL ) {
			//	log_message( "After SetInfo, we did NOT get it" );
			//} else {
			//	log_message1( "After SetInfo, we got", test );
			//}
		} else if ( command->getValue() == FastMode ) {
			timerMode = SuperFast;
		} else if ( command->getValue() == SlowMode ) {
			timerMode = SlowAsMud;
		} else if ( command->getValue() == RotateX ) {
			imv.RotateX();
		} else if ( command->getValue() == RotateY ) {
			imv.RotateY();
		} else if ( command->getValue() == RotateZ ) {
			imv.RotateZ();
		} else if ( command->getValue() == RotateXcc ) {
			imv.RotateXcc();
		} else if ( command->getValue() == RotateYcc ) {
			imv.RotateYcc();
		} else if ( command->getValue() == RotateZcc ) {
			imv.RotateZcc();
		} else if ( command->getValue() == MarkX ) {
			imv.MarkX( command->getParam() );
		} else if ( command->getValue() == MarkY ) {
			imv.MarkY( command->getParam() );
		} else if ( command->getValue() == MarkZ ) {
			imv.MarkZ( command->getParam() );
		} else if ( command->getValue() == SetMarker ) {
			imv.SetMarker();
		} else if ( command->getValue() == CallFunction ) {
			log_message1( "Call:", command->getParam() );
			functionManager.execute( &cq, command->getParam() );
		} else if ( command->getValue() == SpaceX ) {
			imv.SpaceX( command->getParam() );
		} else if ( command->getValue() == SpaceY ) {
			imv.SpaceY( command->getParam() );
		} else if ( command->getValue() == SpaceZ ) {
			imv.SpaceZ( command->getParam() );
		} else if ( command->getValue() == ModelSwitch ) {
			imv.ModelSwitch();
		}
		delete command;
		if ( timerMode == SuperFast ) {
			command = cqp->getNextCommand();
		} else {
			command = NULL;
		}
	}

	CView::OnTimer(nIDEvent);
}


void CCubesView::OnRotateLeft() 
{
	// TODO: Add your command handler code here
	imv.RotateY();
}

void CCubesView::OnRotateRight() 
{
	// TODO: Add your command handler code here
	imv.RotateYcc();
}

void CCubesView::OnRotateTowards() 
{
	// TODO: Add your command handler code here
	imv.RotateX();	
}

void CCubesView::OnRotateAway() 
{
	// TODO: Add your command handler code here
	imv.RotateXcc();
}

void CCubesView::OnTwistLeft() 
{
	// TODO: Add your command handler code here
	imv.RotateZ();
}

void CCubesView::OnTwistRight() 
{
	// TODO: Add your command handler code here
	imv.RotateZcc();
}

void CCubesView::OnJiggle() 
{
	// TODO: Add your command handler code here
	functionManager.execute( &cq, "jiggle" );
}

void CCubesView::OnMoveToMarker() 
{
	// TODO: Add your command handler code here
	imv.MarkX();
//	imv.MarkY();
	imv.MarkZ();
	
}
