/*
   MindtelSocket.cpp

  Minimal socket interface for secure socket data transfer.

  */

#include <stdio.h>
#include <io.h>
#include <winsock2.h>
#include "MindtelSocket.h"

// windows specific garbage
static bool initialized = false;

static int init() {
	if ( !initialized ) {
		WORD wVersionRequested;
		WSADATA wsaData;
		int err;
 
		wVersionRequested = MAKEWORD( 2, 2 );
 
		err = WSAStartup( wVersionRequested, &wsaData );
		if ( err != 0 ) {
			/* Tell the user that we could not find a usable */
			/* WinSock DLL.                                  */
//			printf("WSAStartup didn't fucking work!\n");
			return(-1);
		}
 
		/* Confirm that the WinSock DLL supports 2.2.*/
		/* Note that if the DLL supports versions greater    */
		/* than 2.2 in addition to 2.2, it will still return */
		/* 2.2 in wVersion since that is the version we      */
		/* requested.                                        */
		if ( LOBYTE( wsaData.wVersion ) != 2 ||
			HIBYTE( wsaData.wVersion ) != 2 ) {
			/* Tell the user that we could not find a usable */
			/* WinSock DLL.                                  */
			WSACleanup( );
			return(-1); 
		}
 
	}
	initialized = true;
	return( 0 );
}

static void bzero( char* ptr, int size ) {
	for ( int i = 0; i < size; i++ ) {
		*ptr++ = 0;
	}
}


/*
 *  MindtelSocket::MindtelSocket
 *
 *  Purpose:
 *    Initialize the networking (windows only), then set up the socket.
 */

MindtelSocket::MindtelSocket() {
	if (init() != 0) {
		printf( "Sockets are fucked up on this machine.  Thanks, Microsoft!\n" );
		exit( 0 );
	}

	/* The WinSock DLL is acceptable. Proceed. */
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
}


/*
 *  MindtelSocket::Send
 * 
 *  Purpose:
 *    Send data out connected socket
 *
 *  Input parameters:
 *    buf -- pointer to memory to send out socket
 *    buflen -- how many bytes to send
 */

void MindtelSocket::Send( const char* buf, int buflen ) {
	send( sockfd, buf, buflen, 0 );
}



/*
 *  MindtelSocket::Recv
 * 
 *  Purpose:
 *    Recv data from connected socket
 *
 *  Input parameters:
 *    buf -- pointer to receive destination memory
 *    buflen -- how many bytes to receive
 */

int MindtelSocket::Recv( char* buf, int buflen ) {
	int result = recv( sockfd, buf, buflen, 0 );
	if ( result >= 0 ) {
		*(buf + result) = 0;  // for those callers that are string based
	}
	return( result );
}


/*
 *  MindtelSocketClient::MindtelSocketClient
 *
 *  Purpose:
 *    Constructor for side initiating a socket connection.
 *
 *  Input Parameters:
 *    ip -- ip address of the remote machine
 *    portNumber -- port to connect to
 */

MindtelSocketClient::MindtelSocketClient( char* ipin, int portNo)
: MindtelSocket() {
	memset( ip, 0, 30 );
	strncpy( ip, ipin, 29 );
	portNumber = portNo;
}


/*
 *  MindtelSocketClient::Connect( ip, portNumber ) 
 *
 *  Purpose:
 *    Set up TCP socket connection to a remote machine.
 *
 *  Return Value:
 *    true if the connection succeeds, otherwise false
 */

bool MindtelSocketClient::Connect() {
	struct sockaddr_in servaddr;
	bzero((char*)&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(portNumber);
	servaddr.sin_addr.S_un.S_addr = inet_addr(ip);

    if (connect(sockfd, (const struct sockaddr*) &servaddr, sizeof(servaddr)) == -1) {
		return( false );
	} else {
		return( true );
	}
}


void MindtelSocketClient::GET( char* page ) {
	char buf[500];
	sprintf( buf, "GET %s HTTP/1.0\r\n\r\n", page );
	Send( buf, strlen( buf ));
}
